@@ -18,11 +18,10 @@ import sys
 import types
 from collections import defaultdict, namedtuple
 from itertools import islice
-from lib2to3.pgen2.tokenize import cookie_re as encoding_pattern
 from operator import attrgetter
 from threading import RLock
 
-__version__ = '0.3.2'
+__version__ = '0.4.2'
 __version_info__ = namedtuple(
     'VersionInfo', ('major', 'minor', 'micro')
 )(*(map(int, __version__.split('.'))))
@@ -37,14 +36,20 @@ if PY3:
     # noinspection PyUnresolvedReferences
     from tokenize import detect_encoding
     from itertools import zip_longest
+    # noinspection PyUnresolvedReferences
+    from pathlib import Path
 
     cache = lru_cache(maxsize=None)
     text_type = str
 else:
-    from lib2to3.pgen2.tokenize import detect_encoding
+    from lib2to3.pgen2.tokenize import detect_encoding, cookie_re as encoding_pattern
     from itertools import izip_longest as zip_longest
 
 
+    class Path(object):
+        pass
+
+
     def cache(func):
         d = {}
 
@@ -62,17 +67,19 @@ else:
     text_type = unicode
 try:
     # noinspection PyUnresolvedReferences
-    get_instructions = dis.get_instructions
+    _get_instructions = dis.get_instructions
 except AttributeError:
-    Instruction = namedtuple('Instruction', 'offset argval opname')
+    class Instruction(namedtuple('Instruction', 'offset argval opname starts_line')):
+        lineno = None
 
-    from dis import HAVE_ARGUMENT, EXTENDED_ARG, hasconst, opname
+    from dis import HAVE_ARGUMENT, EXTENDED_ARG, hasconst, opname, findlinestarts
 
     # Based on dis.disassemble from 2.7
     # Left as similar as possible for easy diff
 
-    def get_instructions(co):
+    def _get_instructions(co):
         code = co.co_code
+        linestarts = dict(findlinestarts(co))
         n = len(code)
         i = 0
         extended_arg = 0
@@ -80,6 +87,7 @@ except AttributeError:
             offset = i
             c = code[i]
             op = ord(c)
+            lineno = linestarts.get(i)
             argval = None
             i = i + 1
             if op >= HAVE_ARGUMENT:
@@ -91,7 +99,17 @@ except AttributeError:
 
                 if op in hasconst:
                     argval = co.co_consts[oparg]
-            yield Instruction(offset, argval, opname[op])
+            yield Instruction(offset, argval, opname[op], lineno)
+
+
+def get_instructions(co):
+    lineno = None
+    for inst in _get_instructions(co):
+        lineno = inst.starts_line or lineno
+        assert lineno
+        inst.lineno = lineno
+        yield inst
+
 
 TESTING = 0
 
@@ -129,6 +147,7 @@ class Source(object):
     Attributes:
         - filename
         - text
+        - lines
         - tree: AST parsed from text, or None if text is not valid Python
             All nodes in the tree have an extra `parent` attribute
 
@@ -138,16 +157,21 @@ class Source(object):
         - code_qualname
     """
 
-    def __init__(self, filename, text):
+    def __init__(self, filename, lines):
         """
         Don't call this constructor, see the class docstring.
         """
 
         self.filename = filename
+        text = ''.join(lines)
 
         if not isinstance(text, text_type):
-            text = self.decode_source(text)
+            encoding = self.detect_encoding(text)
+            text = text.decode(encoding)
+            lines = [line.decode(encoding) for line in lines]
+
         self.text = text
+        self.lines = [line.rstrip('\r\n') for line in lines]
 
         if PY3:
             ast_text = text
@@ -158,28 +182,27 @@ class Source(object):
             ast_text = ''.join([
                 '\n' if i < 2 and encoding_pattern.match(line)
                 else line
-                for i, line in enumerate(text.splitlines(True))
+                for i, line in enumerate(lines)
             ])
 
         self._nodes_by_line = defaultdict(list)
         self.tree = None
         self._qualnames = {}
 
-        if text:
-            try:
-                self.tree = ast.parse(ast_text, filename=filename)
-            except SyntaxError:
-                pass
-            else:
-                for node in ast.walk(self.tree):
-                    for child in ast.iter_child_nodes(node):
-                        child.parent = node
-                    if hasattr(node, 'lineno'):
-                        self._nodes_by_line[node.lineno].append(node)
+        try:
+            self.tree = ast.parse(ast_text, filename=filename)
+        except SyntaxError:
+            pass
+        else:
+            for node in ast.walk(self.tree):
+                for child in ast.iter_child_nodes(node):
+                    child.parent = node
+                if hasattr(node, 'lineno'):
+                    self._nodes_by_line[node.lineno].append(node)
 
-                visitor = QualnameVisitor()
-                visitor.visit(self.tree)
-                self._qualnames = visitor.qualnames
+            visitor = QualnameVisitor()
+            visitor.visit(self.tree)
+            self._qualnames = visitor.qualnames
 
     @classmethod
     def for_frame(cls, frame):
@@ -196,8 +219,10 @@ class Source(object):
         except KeyError:
             pass
 
+        if isinstance(filename, Path):
+            filename = str(filename)
         lines = linecache.getlines(filename, module_globals)
-        result = source_cache[filename] = cls(filename, ''.join(lines))
+        result = source_cache[filename] = cls(filename, lines)
         return result
 
     @classmethod
@@ -317,10 +342,14 @@ class Source(object):
     @staticmethod
     def decode_source(source):
         if isinstance(source, bytes):
-            encoding, _ = detect_encoding(io.BytesIO(source).readline)
+            encoding = Source.detect_encoding(source)
             source = source.decode(encoding)
         return source
 
+    @staticmethod
+    def detect_encoding(source):
+        return detect_encoding(io.BytesIO(source).readline)[0]
+
     def code_qualname(self, code):
         """
         Imitates the __qualname__ attribute of functions for code objects.
@@ -353,8 +382,6 @@ class Executing(object):
 
     Generally you will just want `node`, which is the AST node being executed,
     or None if it's unknown.
-    Currently `node` can only be an `ast.Call` object, other operations
-    will be supported in future.
     """
 
     def __init__(self, frame, source, node, stmts):
@@ -444,8 +471,19 @@ class NodeFinder(object):
         self.frame = frame
         self.tree = tree
         self.lasti = lasti
+        self.code = code = frame.f_code
+        self.is_pytest = any(
+            'pytest' in name.lower()
+            for group in [code.co_names, code.co_varnames]
+            for name in group
+        )
 
-        b = frame.f_code.co_code[lasti]
+        if self.is_pytest:
+            self.ignore_linenos = frozenset(assert_linenos(tree))
+        else:
+            self.ignore_linenos = frozenset()
+
+        b = code.co_code[lasti]
         if not PY3:
             b = ord(b)
         op_name = dis.opname[b]
@@ -476,8 +514,16 @@ class NodeFinder(object):
 
             self.result = only(list(self.matching_nodes(exprs)))
 
+    def clean_instructions(self, code):
+        return [
+            inst
+            for inst in get_instructions(code)
+            if inst.opname != 'EXTENDED_ARG'
+            if inst.lineno not in self.ignore_linenos
+        ]
+
     def get_original_instructions(self):
-        hide_instructions = ["EXTENDED_ARG"]
+        result = self.clean_instructions(self.code)
 
         # pypy sometimes (when is not clear)
         # inserts JUMP_IF_NOT_DEBUG instructions in bytecode
@@ -487,12 +533,12 @@ class NodeFinder(object):
                 inst.opname == "JUMP_IF_NOT_DEBUG"
                 for inst in self.compile_instructions()
         ):
-            hide_instructions.append("JUMP_IF_NOT_DEBUG")
+            result = [
+                inst for inst in result
+                if inst.opname != "JUMP_IF_NOT_DEBUG"
+            ]
 
-        return [
-            inst for inst in get_instructions(self.frame.f_code)
-            if inst.opname not in hide_instructions
-        ]
+        return result
 
     def matching_nodes(self, exprs):
         original_instructions = self.get_original_instructions()
@@ -558,18 +604,49 @@ class NodeFinder(object):
                             inst.opname in ('JUMP_FORWARD', 'JUMP_ABSOLUTE')
                             for inst in [inst1, inst2]
                         )
-                ), (inst1, inst2, ast.dump(expr), expr.lineno, self.frame.f_code.co_filename)
+                ), (inst1, inst2, ast.dump(expr), expr.lineno, self.code.co_filename)
 
             yield expr
 
     def compile_instructions(self):
-        module_code = compile_similar_to(self.tree, self.frame.f_code)
-        code = only(find_codes(module_code, self.frame.f_code))
-        return [
-            inst
-            for inst in get_instructions(code)
-            if inst.opname != 'EXTENDED_ARG'
+        module_code = compile_similar_to(self.tree, self.code)
+        code = only(self.find_codes(module_code))
+        return self.clean_instructions(code)
+
+    def find_codes(self, root_code):
+        checks = [
+            attrgetter('co_firstlineno'),
+            attrgetter('co_name'),
+            attrgetter('co_freevars'),
+            attrgetter('co_cellvars'),
         ]
+        if not self.is_pytest:
+            checks += [
+                attrgetter('co_names'),
+                attrgetter('co_varnames'),
+            ]
+
+        def matches(c):
+            return all(
+                f(c) == f(self.code)
+                for f in checks
+            )
+
+        code_options = []
+        if matches(root_code):
+            code_options.append(root_code)
+
+        def finder(code):
+            for const in code.co_consts:
+                if not inspect.iscode(const):
+                    continue
+
+                if matches(const):
+                    code_options.append(const)
+                finder(const)
+
+        finder(root_code)
+        return code_options
 
 
 def get_setter(node):
@@ -589,45 +666,18 @@ def get_setter(node):
 lock = RLock()
 
 
-def find_codes(root_code, matching):
-    def matches(c):
-        return all(
-            f(c) == f(matching)
-            for f in [
-                attrgetter('co_firstlineno'),
-                attrgetter('co_name'),
-                code_names,
-            ]
-        )
-
-    code_options = []
-    if matches(root_code):
-        code_options.append(root_code)
-
-    def finder(code):
-        for const in code.co_consts:
-            if not inspect.iscode(const):
-                continue
-
-            if matches(const):
-                code_options.append(const)
-            finder(const)
-
-    finder(root_code)
-    return code_options
-
-
-def code_names(code):
-    return frozenset().union(
-        code.co_names,
-        code.co_varnames,
-        code.co_freevars,
-        code.co_cellvars,
-    )
-
-
 @cache
 def statement_containing_node(node):
     while not isinstance(node, ast.stmt):
         node = node.parent
     return node
+
+
+def assert_linenos(tree):
+    for node in ast.walk(tree):
+        if (
+                hasattr(node, 'parent') and
+                hasattr(node, 'lineno') and
+                isinstance(statement_containing_node(node), ast.Assert)
+        ):
+            yield node.lineno
